#ifndef Controller_H
#define Controller_H

#include <iostream>
#include <sstream>
#include <list>
#include <string>

#include <QStringListModel>

#include <ros/ros.h>

#include <hatp_msgs/check_database_consistency.h>
#include <hatp_msgs/hatp_msgs.h>

#include <ExtendedTask.hpp>
#include <State.hpp>

namespace hatptester{

class HATPController: public QObject{
	Q_OBJECT

private:
	const std::string htnDatabaseService;
	const std::string factsDatabaseService;
	
	//To remember previous state
	hatp_msgs::WorldBase entitiesMessage;
	std::vector<hatp_msgs::TaskDescription> tasksMessage;

	HATPState state;
	std::map<std::string, ExtendedTask> extendedTasks;
	ExtendedTask* currentTask;
	
	QStringListModel tasksModel;
	
	bool correctlyInitialised;
	ros::NodeHandle handler;
	
	int requestId;
	
	HATPController(): htnDatabaseService("/hatp/htn_database"), factsDatabaseService("/hatp/fact_database"){
		if(!ishatpConnected()) ROS_INFO("HATP is not connected");
		initialiseModel();
	}
	
	HATPController(HATPController const&);
	HATPController& operator=(HATPController const&);
	
	bool hasSameTasks(){
		bool sameDatabase=true;
		ros::ServiceClient client(handler.serviceClient<hatp_msgs::HTNDatabaseRequest>(htnDatabaseService));
		hatp_msgs::HTNDatabaseRequest service;
		
		if(!client.call(service)){
			ROS_INFO("Failed to retrieve the database from service '%s'", client.getService().c_str());
			tasksMessage.clear();
			return false;
		}
		
		std::vector<hatp_msgs::TaskDescription>& tasks=service.response.tasks;
		if(!DatabaseChecker::checkHTNDatabaseConsistency(tasks, false)){
			ROS_ERROR("Invalid HTN database");
			return false;
		}
		
		//Now check if they are the same as the ones stored in "tasksMessage"
		if(tasksMessage.size()!=tasks.size()){
			sameDatabase=false;
		}
		std::vector<hatp_msgs::TaskDescription>::const_iterator it, itDB;
		for(it=tasksMessage.begin(); sameDatabase!=false && it!=tasksMessage.end(); ++it){
			for(itDB=tasks.begin(); sameDatabase!=false && itDB!=tasks.end(); ++itDB){
				if(it->name != itDB->name){
					sameDatabase=false;
					break;
				}
				if(it->type != itDB->type){
					sameDatabase=false;
					break;
				}
				if(it->parameters.size() != itDB->parameters.size()){
					sameDatabase=false;
					break;
				}
				if(it->parameters != itDB->parameters){
					sameDatabase=false;
					break;
				}
			}
		}
		
		if(!sameDatabase){
			tasksMessage=tasks;
			return false;
		}
		return true;
	}

	#define METHODS 0
	#define ACTIONS 1
	//Tasktype describes the type of task to add to the state : either 0 for Methods or 1 for Actions
	bool initialiseTasks(bool taskType){
		if(tasksMessage.empty()) return false;
		
		std::vector<hatp_msgs::TaskDescription>::const_iterator it;
		for(it=tasksMessage.begin(); it!=tasksMessage.end(); ++it){
			//First transform the parameters
			Task::Parameters temp;
			std::vector<std::string>::const_iterator itParam;
			for(itParam=it->parameters.begin(); itParam!=it->parameters.end(); ++itParam){
				std::string name, type;
				name=itParam->substr(0, itParam->find(' '));
				type=itParam->substr(itParam->find(' ')+1, std::string::npos);
				temp.push_back(Task::Parameter(name, type));
			}
			
			if(it->type==TYPE_METHOD && taskType==METHODS){
				if(!state.addMethod(Method(it->name, temp))){
					ROS_ERROR("Error when inserting the task %s in hatptester", it->name.c_str());
					return false;
				}
			}else if(it->type==TYPE_ACTION && taskType==ACTIONS){
				if(!state.addAction(Action(it->name, temp))){
					ROS_ERROR("Error when inserting the entity %s in hatptester", it->name.c_str());
					return false;
				}
			}else continue;
		}
		
		return true;
	}
	
	//Retrieves the entities and check that they are the same as before
	bool hasSameEntities(){
		bool sameDatabase=true;
		ros::ServiceClient client(handler.serviceClient<hatp_msgs::FactDatabaseRequest>(factsDatabaseService));
		hatp_msgs::FactDatabaseRequest service;
		
		if(!client.call(service)){
			ROS_INFO("Failed to call retrieve the database from service '%s'", client.getService().c_str());
			entitiesMessage=hatp_msgs::WorldBase();
			return false;
		}
		
		hatp_msgs::WorldBase& database=service.response.database;
		if(!DatabaseChecker::checkFactDatabaseConsistency(database, false)){
			ROS_ERROR("Invalid fact database");
			return false;
		}
		
		//Now check if they are the same as the ones stored in "entitiesMessage"
		if(entitiesMessage.agent!=database.agent){
			sameDatabase=false;
		}
		if(entitiesMessage.entities.size()!=database.entities.size()){
			sameDatabase=false;
		}
		for(std::vector<hatp_msgs::Entity>::const_iterator itEnt=entitiesMessage.entities.begin(); sameDatabase!=false && itEnt!=entitiesMessage.entities.end(); ++itEnt){
			for(std::vector<hatp_msgs::Entity>::const_iterator itEntDB=database.entities.begin(); sameDatabase!=false && itEntDB!=database.entities.end(); ++itEntDB){
				if(itEnt->name!=itEntDB->name) continue;
				
				if(itEnt->type!=itEntDB->type){
						sameDatabase=false;
						break;
				}
				if(itEnt->attributes.size()!=itEntDB->attributes.size()){
					sameDatabase=false;
					break;
				}
				
				for(std::vector<hatp_msgs::Attribute>::const_iterator itAttr=itEnt->attributes.begin(); itAttr!=itEnt->attributes.end(); ++itAttr){
					for(std::vector<hatp_msgs::Attribute>::const_iterator itAttrDB=itEntDB->attributes.begin(); itAttrDB!=itEntDB->attributes.end(); ++itAttrDB){
						if(itAttr->name!=itAttrDB->name) continue;
						
						if(itAttr->type!=itAttrDB->type){
							sameDatabase=false;
							break;
						}
						if(itAttr->dynamic!=itAttrDB->dynamic){
							sameDatabase=false;
							break;
						}
						if(itAttr->set!=itAttrDB->set){
							sameDatabase=false;
							break;
						}
						if(itAttr->values.size()!=itAttrDB->values.size()){
							sameDatabase=false;
							break;
						}
						if(itAttr->values!=itAttrDB->values){
							sameDatabase=false;
							break;
						}
					}
				}
			}
		}

		if(sameDatabase==false){
			entitiesMessage=database;
		}
		return sameDatabase;
	}
	
	bool initialiseEntities(){
		if(entitiesMessage.entities.empty()) return false;
		
		for(std::vector<hatp_msgs::Entity>::const_iterator itEnt=entitiesMessage.entities.begin(); itEnt!=entitiesMessage.entities.end(); ++itEnt){
			if(!state.addEntity(Entity(itEnt->type, itEnt->name))){
				ROS_ERROR("Error when inserting the entity %s in hatptester", itEnt->name.c_str());
				return false;
			}
		}
		
		return true;
	}
	
	void initialiseTaskModel(){
		QStringList tasks;
		//TODO Remove those marking and put unselectable labels instead (with different backgroung color)
		tasks << " -- Methods : ";
		
		for(std::map<std::string, Method>::const_iterator it=state.getMethods().begin(); it!=state.getMethods().end(); ++it){
			tasks << it->first.c_str();
			extendedTasks.insert(std::make_pair(it->first, ExtendedTask(state, it->first)));
		}
		tasks << "";

		tasks << " -- Actions : ";
		for(std::map<std::string, Action>::const_iterator it=state.getActions().begin(); it!=state.getActions().end(); ++it){
			tasks << it->first.c_str();
			extendedTasks.insert(std::make_pair(it->first, ExtendedTask(state, it->first)));
		}

		tasksModel.setStringList(tasks);
	}
	
	void clearTaskModel(){
		QStringList tasks;
		tasksModel.setStringList(tasks);
	}
	
	bool initialiseModel(){
		bool success=false;
		correctlyInitialised=false;
		bool changed=false;
		
		if(ishatpConnected()){
			//Initialise state using requests to HATP
			if(!hasSameEntities()){
				state.clearEntities();
				success=initialiseEntities();
				changed=true;
			} else success=true;

			if(success){
				if(!hasSameTasks()){
					state.clearMethods();
					success=initialiseTasks(METHODS); //Get Methods
					state.clearActions();
					success=initialiseTasks(ACTIONS); //Get Actions
					changed=true;
				} else success=true;
					
				if(success){		
					if(changed){//Initialise the different models for Qt GUI
						extendedTasks.clear();
						initialiseTaskModel();
						
						currentTask=NULL;
					}
					requestId=0;
					correctlyInitialised=true;
				}else{
					ROS_WARN("Error when retrieveing the tasks");
				}
			}else{
				ROS_WARN("Error when retrieveing the entities");
			}
		}else{
			ROS_ERROR("Client failed to connect");
			return false;
		}
		
		if(!success){
			ROS_WARN("Failed to initialise the model from the database");
			state.clearEntities();
			state.clearMethods();
			state.clearActions();
			clearTaskModel();
			entitiesMessage=hatp_msgs::WorldBase();
			tasksMessage.clear();
		}
// 		if(changed){
// 			emit sendStateChanged();
// 		}
		return success;
	}
	
	void sendRequest(bool stopAtFirstPlan){
		if(currentTask==NULL) return;
		
		//True means it must stop after the first valid plan found
		//False means it computes all possible plans
		ros::ServiceClient client = handler.serviceClient<hatp_msgs::PlanningRequest>("/hatp/planner");
		
		hatp_msgs::PlanningRequest service;
		service.request.request.type=(stopAtFirstPlan?"first-plan":"plan");
		service.request.request.id = requestId++;
		
		service.request.request.task=currentTask->getName();
		
		std::pair<bool, const Task*> result;
		result=state.getTask(currentTask->getName());
		if(!result.first){
			return;
		}
		const Task::Parameters& parameters=result.second->getParameters();
		for(Task::Parameters::const_iterator it=parameters.begin(); it!=parameters.end(); ++it){
			service.request.request.parameters.push_back(it->value);
		}
		
		if(!client.call(service)){
			ROS_WARN("Failed to call service 'Planner'");
			return;
		}
	}
	
public:	
	static HATPController& getInstance(){
		static HATPController instance;
		return instance;
	}
	
	bool isConnected(){
		return handler.ok();
	}
	
	bool ishatpConnected(){
		return isConnected() && ros::service::exists(htnDatabaseService, false) && ros::service::exists(factsDatabaseService, false);
	}
	
	bool isReady(){
		return(ishatpConnected()&&correctlyInitialised);
	}
	
	void reloadModel(){
		initialiseModel();
	}
	
	const std::map<std::string, Entity> getEntities() const{
		return(state.getEntities());
	}
	
	void sendFirstPlanRequest(){
		sendRequest(true); //True means it stops after the first valid plan found
	}
	
	void sendAllPlansRequest(){
		sendRequest(false); //False means it computes all possible plans
	}
	
	void sendMaxTime(int value){ //value<=0 means to send an unlimited planning time, any other value means limited planning time
		std::cout<<"SEND MAX TIME REQUEST !!! - NOT IMPLEMENTED YET"<<std::endl;
// 		std::stringstream request;
// 		request<<"{\"HATP-REQ\": { \"REQ-TYPE\": ";
// 		
// 		if(value<=0){
// 			request<<"\"unsetMaxTime\"";
// 		}else{
// 			request<<"\"setMaxTime\", \"MAXTIME\": "<<value;
// 		}
// 		
// 		request<<"}}";
// 		
// 		LOG(trace)<<"Send the request : "<<request.str();
// 		
// 		//Send the request
// 		if(!client.sendMessage(request.str())){
// 			LOG(error)<<"Error while sending the message";
// 			this->exit();
// 		}
	}
	
	QStringListModel& getTasks(){
		return tasksModel;
	}
	
	QStandardItemModel* getTaskParameter(const std::string& task){
		std::map<std::string, ExtendedTask>::iterator result;
		result=extendedTasks.find(task);
		if(result!=extendedTasks.end()){
			currentTask=&result->second;
			return result->second.getParametersModel();
		}
		return NULL;
	}
	
	bool isReadyTask(const std::string& task){
		std::pair<bool, const Task*> result;
		result=state.getTask(task);
		if(result.first){
			return result.second->isReady();
		}
		return false;
	}
	
	bool isReadyCurrentTask(){
		if(currentTask==NULL) return false;
		
		std::pair<bool, const Task*> result;
		result=state.getTask(currentTask->getName());
		if(result.first){
			return result.second->isReady();
		}
		return false;
	}
	
	const std::string getCurrentTaskName() const{
		if(currentTask==NULL) return "";
		
		return currentTask->getName();
	}
	
	std::set<std::string> getMissingParameters(const std::string& task){
		std::set<std::string> missingParameters;
	
		std::pair<bool, const Task*> result;
		result=state.getTask(task);
		if(result.first && !result.second->isReady()){
			const Task::Parameters& parameters=result.second->getParameters();
			for(Task::Parameters::const_iterator it=parameters.begin(); it!=parameters.end(); ++it){
				if(it->value==""){
					missingParameters.insert(it->name+" ["+it->type+"]");
				}
			}
		}
		return missingParameters;
	}
	
	const std::string getLongestTaskName() const{
		std::string longest="";
		for(HATPState::MethodsMap::const_iterator it=state.getMethods().begin(); it!=state.getMethods().end(); ++it){
			if(it->first.length()>longest.length()){
				longest=it->first;
			}
		}
		for(HATPState::ActionsMap::const_iterator it=state.getActions().begin(); it!=state.getActions().end(); ++it){
			if(it->first.length()>longest.length()){
				longest=it->first;
			}
		}
		return longest;
	}
	
signals:
	void sendStateChanged();
};

}; //namespace

#endif //Controller_H;
